#include "../../include/ctx.h"
#include "../../include/pool.h"
#include "../../include/util.h"
#include "../../include/error.h"
#include "../../include/thpool.h"

#include <errno.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

lm_pool_t *lm_ctx_pools_add(lm_ctx_t *ctx, char *name, char *url) {
  lm_pool_t *pool = lm_pool_new(name, url);
  if (NULL == pool)
    return false;

  pdebug(__func__, "pool name is %s", pool->name);
  pdebug(__func__, "pool URL is %s://%s:%d%s", pool->url.protocol, pool->url.host, pool->url.port, pool->url.path);

  if (NULL == ctx->pools) {
    ctx->pools = pool;
    return pool;
  }

  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (NULL == cur->next) {
      cur->next = pool;
      return pool;
    }
    cur = cur->next;
  }

  lm_pool_free(pool);
  return NULL;
}

void lm_ctx_pools_clear(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools, *prev = NULL;
  while (NULL != cur) {
    prev = cur;
    cur  = cur->next;
    lm_pool_free(prev);
  }

  ctx->pools = NULL;
}

void lm_ctx_pools_test(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    lm_pool_test(cur);
    if (!cur->available)
      pdebug(__func__, "%s is not avaliable: %s", cur->name, lm_strerror());
    cur = cur->next;
  }
}

void lm_ctx_pools_get_info(lm_ctx_t *ctx) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    lm_pool_test(cur);
    if (!cur->available)
      pdebug(__func__, "%s is not avaliable: %s", cur->name, lm_strerror());
    cur = cur->next;
  }
}

lm_pool_t *lm_ctx_pools_find(lm_ctx_t *ctx, char *name) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (eq(cur->name, name))
      return cur;
    cur = cur->next;
  }
  return NULL;
}

lm_pool_t *lm_ctx_pools_by_url(lm_ctx_t *ctx, char *host, char *path) {
  lm_pool_t *cur = ctx->pools;
  while (NULL != cur) {
    if (eq(cur->url.host, host) && eq(cur->url.path, path))
      return cur;
    cur = cur->next;
  }
  return NULL;
}

bool lm_ctx_pool_load(lm_ctx_t *ctx, lm_pool_t *pool, bool force_update){
  if(NULL == ctx->data){
    printf("hello chat\n");
    lm_error_set(LM_ERR_CtxDataNULL);
    return false;
  }

  char poolp[strlen(ctx->data) + strlen(pool->name) + 20];
  snprintf(poolp, sizeof(poolp), "%s/pools/%s", ctx->data, pool->name);
  size_t poolp_sz = strlen(poolp);

  if(!mkdir_ifnot(poolp)){
    lm_error_set(LM_ERR_CtxDataFailMkdir);
    return false;
  }

  char infop[poolp_sz + 10];
  sprintf(infop, "%s/INFO", poolp);
  
  char listp[poolp_sz + 10];
  sprintf(listp, "%s/LIST", poolp);

  if(force_update)
    goto update;

  if(!lm_pool_info_load(pool, infop)){
    pdebug(__func__, "failed to load info for %s gonna try updating", pool->name); 
    goto update;
  }

  if(!lm_pool_info_load(pool, listp)){
    pdebug(__func__, "failed to load list for %s gonna try updating", pool->name); 
    goto update;
  }

update:
  unlink(infop);
  unlink(listp);

  if(!pool->available){
    pdebug(__func__, "failed get info and list for %s, pool is not avaliable");
    return false;
  }

  if(!lm_pool_info_get(pool, infop)){
    pdebug(__func__, "failed to get info for %s", pool->name); 
    return false;
  }

  if(!lm_pool_list_get(pool, listp)){
    pdebug(__func__, "failed to get list for %s", pool->name); 
    return false;
  }

  return true;
}

bool lm_ctx_pools_load(lm_ctx_t *ctx, bool force_update, lm_ctx_pools_callback_t callback, void *data){
  if(NULL == ctx){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  lm_pool_t *cur = ctx->pools;
  while(NULL != cur){
    bool status = lm_ctx_pool_load(ctx, cur, force_update);

    if(!status)
      pdebug(__func__, "failed to load pool: %s", cur->name);

    if(NULL != callback)
      callback(ctx, cur, status, data);

    cur = cur->next;
  }

  return true;
}

bool lm_ctx_pools_serve(lm_ctx_t *ctx, char *addr, uint8_t threads) {
  if (NULL == addr || threads < 0) {
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  struct sockaddr saddr;
  char           *host = NULL;
  lm_mptp_t       packet;
  uint16_t        port;
  int             sock;
  lm_thpool_t     tp;

  lm_thpool_init(&tp, threads);

  if (!parse_host(addr, host, &port))
    return false;

  if (port == 0) {
    lm_error_set(LM_ERR_NoPort);
    return false;
  }

  if ((sock = lm_mptp_server_listen(addr, port)) < 0)
    return false;

  while (lm_mptp_server_recv(sock, &packet, &saddr)) {
    if (!lm_mptp_server_verify(&packet)) {
      pdebug(__func__, "skipping packet, failed to verify: %s", lm_strerror());
      continue;
    }

    char hostname[packet.header.host_size + 1]; // +1 for NULL terminator
    char path[packet.header.data_size + 1];

    if (!lm_mptp_get_data(&packet, path)) {
      pdebug(__func__, "skipping packet, failed to get path: %s", lm_strerror());
      continue;
    }

    if (!lm_mptp_get_host(&packet, hostname)) {
      pdebug(__func__, "skipping packet, failed to get hostname: %s", lm_strerror());
      continue;
    }

    lm_pool_t *pool = lm_ctx_pools_by_url(ctx, hostname, path);

    if (NULL == pool) {
      pdebug(__func__, "unknown pool, closing connection: %s", hostname);
      lm_mptp_init(&packet, false, MPTP_S2C_BRUH, true);
      lm_mptp_server_send(sock, &packet, &saddr);
      continue;
    }

    lm_pool_thread_arg_t *arg = malloc(sizeof(lm_pool_thread_arg_t));

    memcpy(&arg->addr, &saddr, sizeof(struct sockaddr));
    lm_mptp_copy(&arg->packet, &packet);
    arg->pool = pool;
    arg->sock = sock;

    pdebug(__func__, "adding new connection to the pool");
    lm_thpool_add(&tp, lm_pool_serve_thread, arg);
  }

  lm_thpool_stop(&tp);
  return true;
}
